/*
 * (c) 2010-2018, SWD Embedded Systems Limited, http://www.kpda.ru
 */


/*
#ifdef __USAGE
%C - Serial driver for 8250's

%C [options] [port[^shift][,irq]] &
Options:
 -b number    Define initial baud rate (default 57600)
 -c clk[/div] Set the input clock rate and divisor
 -C number    Size of canonical input buffer (default 256)
 -e           Set options to "edit" mode
 -E           Set options to "raw" mode (default)
 -I number    Size of raw input buffer (default 2048)
 -f           Enable hardware flow control (default)
 -F           Disable hardware flow control
 -O number    Size of output buffer (default 2048)
 -s           Enable software flow control
 -S           Disable software flow control (default)
 -t number    Set receive FIFO trigger level (default 8)
 -T number    Set transmit FIFO size (default 14)
 -u unit      Set serial unit number (default 1)
 -m           Map device physical memory instead of I/O memory (x86)
 -o opt[,opt] string options:
              pmm_parent=path (pmm_parent power pathname, default NULL)
              pm_noflw (Disable flow control on power down, default enabled)
              highwater=value (RX watermark for input flow control (bytes))
              disable=rx (Disable receiver on startup)
              timer_period=value (Internal timer period, default 50ms)
 -V 0xXXXX    PCI UART vendor
 -D 0xXXXX    PCI device
 -i idx       PCI index (default 0)
 -v           Increase verbosity
 
Examples:
Autoscan for serial devices from startup system page,
 PCI bus and BIOS default ports (x86)
%C -v -b115200

Start serial driver for default x86 ports
%C -b115200 3f8,4 2f8,3

Start serial driver for Intel PCI HSUART
%C -v -b115200 -V 0x8086 -D 0x0f0a
#endif
*/


#include "externs.h"
#include "hw/sysinfo.h"
#include "drvr/hwinfo.h"


/* List of supported PCI devices */
struct pci_list_t pci_list[] = { {BTUART_PCI_VENDOR, BTUART_PCI_DEVICE_0}, /* Intel Atom E38xx Baytrail HSUART*/
                                 {BTUART_PCI_VENDOR, BTUART_PCI_DEVICE_1},
                                 {0, 0} };


int query_hwi_device( TTYINIT *dip, unsigned unit )
{
    unsigned    hwi_off = hwi_find_device( "uart", unit );

    if ( hwi_off != HWI_NULL_OFF )
    {
        hwi_tag *tag_location = hwi_tag_find( hwi_off, HWI_TAG_NAME_location, 0 );

        if ( tag_location )
        {
            dip->port       = tag_location->location.base;
            dip->port_shift = tag_location->location.regshift;
        }
        hwi_tag *tag_irq = hwi_tag_find( hwi_off, HWI_TAG_NAME_irq, 0 );
        if ( tag_irq )
            dip->intr = tag_irq->irq.vector;

        return (1);
    }

    /* No default device, the base address and irq have been specified */
    return (0);
}


unsigned options( int argc, char *argv[] )
{
    int                 opt,
                        i,
                        enabled_cnt      = 0,
                        numports         = 0,
                        map_dev_mem      = 0;
    char                *cp;
    void                *link;
    unsigned            unit;
    unsigned            fifo_tx, fifo_rx;
    int                 found_hwi_device = -1;
    DEV_8250            *dev_list[MAX_DEVICES];
    static TTYINIT      devinit          = { 0, 0, 0, 57600,
                                             2048, 2048, 256,
                                             0, 0, 0, 0, 0xe8, 0, 0,    /* Default Tx fifo 14, Rx fifo 8 */
                                             "/dev/ser",
                                             NULL, 0, 0, 0     };
    int32_t             pci_id,
                        pci_idx          = 0;
    struct pci_dev_info pci_inf;
    void                *pci_handle      = NULL;
    off64_t             offset;

    pci_inf.DeviceId = pci_inf.VendorId = 0;

    /* Initialize the devinit to raw mode */
    ttc( TTC_INIT_RAW, &devinit, 0 );

    sys_ttyinit( &devinit );
    unit = 1;
    while ( optind < argc )
    {
        /* Process dash options. */
        while ( (opt = getopt( argc, argv, IO_CHAR_SERIAL_OPTIONS "c:t:T:u:mi:V:D:" )) != -1 )
        {
            switch ( ttc( TTC_SET_OPTION, &devinit, opt ) )
            {
                case 'c':
                    devinit.clk = strtoul( optarg, &optarg, 0 );
                    if ( (cp = strchr( optarg, '/' )) )
                        devinit.div = strtoul( cp + 1, NULL, 0 );
                    break;

                case 't':
                    fifo_rx = strtoul( optarg, NULL, 0 );
                    if ( !((fifo_rx == 1) || (fifo_rx == 4) || (fifo_rx == 8) || (fifo_rx == 14)) )
                    {
                        fprintf( stderr, "Illegal rx fifo trigger. \n" );
                        fprintf( stderr, "Trigger number must be 1, 4, 8 or 14. \n" );
                        fprintf( stderr, "Rx trigger will not be enabled.\n\n" );
                        fifo_rx = 0;
                    } else {
                        devinit.fifo &= 0xf0;
                        devinit.fifo |= fifo_rx;
                    }
                    break;

                case 'T':
                    fifo_tx = strtoul( optarg, NULL, 0 );
                    if ( !((fifo_tx == 1) || (fifo_tx == 4) || (fifo_tx == 8) || (fifo_tx == 14)) )
                    {
                        fprintf( stderr, "Illegal tx fifo size. \n" );
                        fprintf( stderr, "Tx fifo size must be 1, 4, 8 or 14. \n" );
                        fprintf( stderr, "Tx fifo will not be enabled.\n\n" );
                        fifo_tx = 0;
                    } else {
                        devinit.fifo &= 0x0f;
                        devinit.fifo |= (fifo_tx << 4);
                    }
                    break;

                case 'u':
                    unit = strtoul( optarg, NULL, 0 );
                    break;

                case 'm':
                    map_dev_mem = 1;
                    break;

                /* PCI index */
                case 'i':
                    pci_idx = strtoul( optarg, NULL, 0 );
                    break;

                /* PCI Vendor */
                case 'V':
                    pci_inf.VendorId = strtoul( optarg, NULL, 0 );
                    break;

                /* PCI device */
                case 'D':
                    pci_inf.DeviceId = strtoul( optarg, NULL, 0 );
                    break;
            }
        }

        /* Process ports and interrupts. */
        while ( optind < argc  &&  *(optarg = argv[optind]) != '-' )
        {
            devinit.port = strtoull( optarg, &optarg, 16 );
            if ( *optarg == '^' )
                devinit.port_shift = strtoul( optarg + 1, &optarg, 0 );

            if ( *optarg == ',' )
                devinit.intr = strtoul( optarg + 1, NULL, 0 );

            if ( (dev_list[numports] = create_device( &devinit, unit++, map_dev_mem, 0 )) == NULL )
            {
                slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
                fprintf( stderr, "io-char: Initialization of port 0x%llx failed\n", devinit.port );
            } else
                ++numports;

            ++optind;
        }
    }

    /* PCI UART device if specified */
    if ( pci_inf.VendorId && pci_inf.DeviceId )
    {
        if ( (pci_id = pci_attach( 0 )) == -1 )
        {
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Unable to attach to PCI server (%d)", errno );
            fprintf( stderr, "io-char: Unable to attach to PCI server (%d)\n", errno );

            return (0);
        }

        if ( (pci_handle = pci_attach_device( NULL, PCI_SEARCH_VENDEV | PCI_MASTER_ENABLE | PCI_INIT_ALL, pci_idx, &(pci_inf) )) == NULL )
        {
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Can't attach to PCI device 0x%X:0x%X", pci_inf.VendorId, pci_inf.DeviceId );
            fprintf( stderr, "io-char: Can't attach to PCI device 0x%X:0x%X\n", pci_inf.VendorId, pci_inf.DeviceId );

            return (0);
        }

        devinit.port       = PCI_MEM_ADDR( pci_inf.CpuBaseAddress[0] );
        devinit.intr       = pci_inf.Irq;
        devinit.port_shift = 2;
        map_dev_mem        = 1;

        if ( (dev_list[numports] = create_device( &devinit, unit++, map_dev_mem, pci_inf.DeviceId )) == NULL )
        {
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
            fprintf( stderr, "io-char: Initialization of port 0x%llx failed\n", devinit.port );

            return (0);
        } else
            ++numports;
    }

    if ( numports == 0 )
    {
        unit = 0;
        link = NULL;

        /* Scan syspage for UART devices */
        for ( ;; )
        {
            found_hwi_device = query_hwi_device( &devinit, unit );
            if ( !found_hwi_device )
                break;

            /* Getting the UART Clock from the Hwinfo Section if available */
            {
                unsigned hwi_off = hwi_find_device( "uart", 0 );
                if ( hwi_off != HWI_NULL_OFF )
                {
                    hwi_tag *tag = hwi_tag_find( hwi_off, HWI_TAG_NAME_inputclk, 0 );
                    if ( tag )
                        devinit.clk = ((tag->inputclk.clk) / (tag->inputclk.div));
                }
            }

            if ( (dev_list[numports] = create_device( &devinit, unit++, map_dev_mem, 0 )) == NULL )
            {
                slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
                fprintf( stderr, "io-char: Initialization of port 0x%llx failed\n", devinit.port );
            } else
                ++numports;
        }
        
        /* Default UART ports if present */
        for ( ;; )
        {
            link = query_default_device( &devinit, link );
            if ( link == NULL )
                break;
            if ( (dev_list[numports] = create_device( &devinit, unit++, map_dev_mem, 0 )) == NULL )
            {
                slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
                fprintf( stderr, "io-char: Initialization of port 0x%llx failed\n", devinit.port );
            } else
                ++numports;
        }

        /* Scan for supported PCI UART devices */
        if ( (pci_id = pci_attach( 0 )) != -1 )
        {
            for ( i = 0; (pci_list[i].vendorID != 0 && pci_list[i].deviceID != 0); i++ )
            {
                pci_inf.VendorId = pci_list[i].vendorID;
                pci_inf.DeviceId = pci_list[i].deviceID;
                if ( (pci_handle = pci_attach_device( NULL, PCI_SEARCH_VENDEV | PCI_MASTER_ENABLE | PCI_INIT_ALL, pci_idx, &(pci_inf) )) == NULL )
                {
                    slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Can't attach to PCI device 0x%X:0x%X", pci_inf.VendorId, pci_inf.DeviceId );
                    fprintf( stderr, "io-char: Can't attach to PCI device 0x%X:0x%X\n", pci_inf.VendorId, pci_inf.DeviceId );

                    return (0);
                }

                devinit.port       = PCI_MEM_ADDR( pci_inf.CpuBaseAddress[0] );
                devinit.intr       = pci_inf.Irq;
                devinit.port_shift = 2;
                map_dev_mem        = 1;

                if ( (dev_list[numports] = create_device( &devinit, unit++, map_dev_mem, pci_inf.DeviceId )) == NULL )
                {
                    slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_ERROR, "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
                    fprintf( stderr, "io-char: Initialization of port 0x%llx failed\n", devinit.port );

                    return (0);
                } else
                    ++numports;
            }
        }
    }

    /* Enable all ports (Attach ISR and pathname entry) */ 
    for ( i = 0; i < numports; i++ )
    {
        if ( dev_list[i]->tty.verbose )
        {
            if ( dev_list[i]->dev_id )
                mem_offset64( (void *)dev_list[i]->port[0], NOFD, 1, &offset, 0 );
            else
                offset = dev_list[i]->port[0];

            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Port .......................... %s (0x%llx)", dev_list[i]->tty.name, offset );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "IRQ ........................... 0x%x", dev_list[i]->intr );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input Clock ................... %d", dev_list[i]->clk );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Clock Divisor ................. %d", dev_list[i]->div );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Tx fifo size .................. %d", (dev_list[i]->tty.fifo >> 4) );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Rx fifo trigger ............... %d", (dev_list[i]->tty.fifo & 0x0f) );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input buffer size ............. %d", dev_list[i]->tty.ibuf.size );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input flow control highwater .. %d", dev_list[i]->tty.highwater );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Output buffer size ............ %d", dev_list[i]->tty.obuf.size );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Canonical buffer size ......... %d\n", dev_list[i]->tty.cbuf.size );
        }

        if ( enable_device( dev_list[i] ) != -1 )
            enabled_cnt++;
    }

    return (enabled_cnt);
}
